from rich.console import Console
from rich.progress import Progress, SpinnerColumn, TextColumn
from contextlib import contextmanager
import sys


@contextmanager
def progress_context(
    description: str, total: int = 9999, transient: bool = True
):
    console = Console(file=sys.stderr)  # Direct output to standard error
    with Progress(
        SpinnerColumn(),
        TextColumn("[progress.description]{task.description}"),
        console=console,  # Use the custom console
        transient=transient,
    ) as progress:
        progress.add_task(description=description, total=total)
        yield



from pydantic import BaseModel, Field
from typing import Optional, List


class Golden(BaseModel):
    input: str
    actual_output: Optional[str] = Field(None, alias="actualOutput")
    expected_output: Optional[str] = Field(None, alias="expectedOutput")
    context: Optional[list] = Field(None)


class APIDataset(BaseModel):
    alias: str
    goldens: Optional[List[Golden]] = Field(default=None)


class CreateDatasetHttpResponse(BaseModel):
    link: str



"""Simple Retry class
"""
import time
from functools import wraps
from typing import Callable, List


class retry:
    retry_count: int = 0
    success_count: int = 0

    def __init__(
        self,
        max_retries: int = 3,
        delay: int = 1,
        min_success: int = 1,
        fail_hooks: List[Callable] = None,
    ) -> None:
        self.max_retries = max_retries
        self.delay = delay
        self.min_success = min_success
        self.fail_hooks = fail_hooks if fail_hooks else []

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            last_error = None  # Store the last error
            while self.retry_count < self.max_retries:
                try:
                    self.retry_count += 1
                    for _ in range(self.min_success - self.success_count):
                        result = func(*args, **kwargs)
                        self.success_count += 1
                except AssertionError as e:
                    print(
                        f"Attempt {self.retry_count} to execute {func.__name__} failed due to AssertionError: {str(e)}"
                    )
                    last_error = e  # Update the last error
                    for fail_hook in self.fail_hooks:
                        fail_hook(e)
                except Exception as e:
                    print(
                        f"Attempt {self.retry_count} to execute {func.__name__} failed due to Exception: {str(e)}"
                    )
                    last_error = e  # Update the last error
                    for fail_hook in self.fail_hooks:
                        fail_hook(e)

                if self.retry_count < self.max_retries:
                    print(
                        f"Retrying execution of {func.__name__} in {self.delay} seconds..."
                    )
                    time.sleep(self.delay)
                else:
                    if self.retry_count != 1:
                        # No need to print it out if you can only retry once
                        print(
                            f"Max retries ({self.max_retries}) for executing {func.__name__} exceeded."
                        )
                    if self.success_count < self.min_success:
                        if last_error is not None:
                            raise last_error  # Raise the last error

        return wrapper


def retry_with_input_modification(self, func, modify_input):
    """
    This function allows you to retry a function with modified input if it fails.

    Example:
    ```python
    def modify_input(*args, **kwargs):
        # Modify the input here
        # In this example, we add a word to the input string
        args = tuple(arg + " word" for arg in args)
        return args, kwargs

    @retry_with_input_modification(modify_input=modify_input)
    def test_function(input_string):
        # Function implementation here
    ```
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        last_error = None  # Store the last error
        while self.retry_count < self.max_retries:
            try:
                self.retry_count += 1
                for _ in range(self.min_success - self.success_count):
                    result = func(*args, **kwargs)
                    self.success_count += 1
            except AssertionError as e:
                print(
                    f"Attempt {self.retry_count} to execute {func.__name__} failed due to AssertionError: {str(e)}"
                )
                last_error = e  # Update the last error
                for fail_hook in self.fail_hooks:
                    fail_hook(e)
                args, kwargs = modify_input(*args, **kwargs)
            except Exception as e:
                print(
                    f"Attempt {self.retry_count} to execute {func.__name__} failed due to Exception: {str(e)}"
                )
                last_error = e  # Update the last error
                for fail_hook in self.fail_hooks:
                    fail_hook(e)
                args, kwargs = modify_input(*args, **kwargs)

            if self.retry_count < self.max_retries:
                print(
                    f"Retrying execution of {func.__name__} in {self.delay} seconds..."
                )
                time.sleep(self.delay)
            else:
                if self.retry_count != 1:
                    # No need to print it out if you can only retry once
                    print(
                        f"Max retries ({self.max_retries}) for executing {func.__name__} exceeded."
                    )
                if self.success_count < self.min_success:
                    if last_error is not None:
                        raise last_error  # Raise the last error

    return wrapper



import torch
from rouge_score import rouge_scorer
from nltk.tokenize import word_tokenize
from nltk.translate.bleu_score import sentence_bleu
from typing import Union, List, Optional, Any
from deepeval.utils import normalize_text
from deepeval.models.summac_model import SummaCZS


# TODO: More scores are to be added
class Scorer:
    """This class calculates various Natural Language Processing (NLP) evaluation score.

    The scoring logic can be a simple algorithm or any statistical formula. There are some scores
    Which also uses an external model (BERTScore) in the scoring logic.
    """

    @classmethod
    def rouge_score(
        cls, target: str, prediction: str, score_type: str
    ) -> float:
        """Calculates the Rouge score for a given target and prediction.

        Rouge (Recall-Oriented Understudy for Gisting Evaluation) is a metric used for evaluating the quality of generated text,
        especially in tasks like text summarization.

        Args:
            target (str): The actual label or target text.
            prediction (str): The generated text from the model or LLM.
            score_type (str): The Rouge score type (Options: 'rouge1', 'rouge2', 'rougeL').

        Returns:
            float: The Rouge score for the given target and prediction, based on the specified score type.
        """
        assert score_type in [
            "rouge1",
            "rouge2",
            "rougeL",
        ], "score_type can be either rouge1, rouge2 or rougeL"
        scorer = rouge_scorer.RougeScorer([score_type], use_stemmer=True)
        scores = scorer.score(target, prediction)
        return scores[score_type].fmeasure

    @classmethod
    def sentence_bleu_score(
        cls,
        references: Union[str, List[str]],
        prediction: str,
        bleu_type: Optional[str] = "bleu1",
    ) -> float:
        """Calculates the BLEU (Bilingual Evaluation Understudy) score for a given prediction compared to one or more reference sentences.

        BLEU is a metric used to evaluate the quality of machine-generated text by comparing it to one or more reference sentences.
        It measures the similarity of the generated text to the reference text based on n-grams.

        Args:
            references (Union[str, List[str]): A reference sentence or a list of reference sentences.
            prediction (str): The generated text or sentence to be evaluated.
            bleu_type (Optional[str]): The BLEU score type (Options: 'bleu1', 'bleu2', 'bleu3', 'bleu4'). Default is 'bleu1'.

        Returns:
            float: The BLEU score for the given prediction and references.
        """
        assert bleu_type in [
            "bleu1",
            "bleu2",
            "bleu3",
            "bleu4",
        ], "Invalud bleu_type. Options: 'bleu1', 'bleu2', 'bleu3', 'bleu4'"
        targets = [references] if isinstance(references, str) else references
        tokenized_targets = [word_tokenize(target) for target in targets]
        tokenized_prediction = word_tokenize(prediction)
        bleu_weight_map = {
            "bleu1": (1, 0, 0, 0),
            "bleu2": (0, 1, 0, 0),
            "bleu3": (0, 0, 1, 0),
            "bleu4": (0, 0, 0, 1),
        }
        return sentence_bleu(
            tokenized_targets,
            tokenized_prediction,
            weights=bleu_weight_map[bleu_type],
        )

    @classmethod
    def exact_match_score(cls, target: str, prediction: str) -> int:
        """Metrics that calculates whether two sequences matches exactly or not.

        Args:
            target (str): The target string.
            prediction (str): The predicted string from the llm

        Returns:
            int: The exact match score.
        """
        if not prediction:
            return 0
        return 1 if prediction.strip() == target.strip() else 0

    @classmethod
    def quasi_exact_match_score(cls, target: str, prediction: str) -> int:
        if not prediction:
            return 0
        return 1 if normalize_text(target) == normalize_text(prediction) else 0

    # Todo: More mode based metrics to be added

    @classmethod
    def bert_score(
        cls,
        references: Union[str, List[str]],
        predictions: Union[str, List[str]],
        model: Optional[str] = "microsoft/deberta-large-mnli",
        lang: Optional[str] = "en",
    ) -> float:
        """
        Calculate BERTScore for one or more reference sentences compared to one or more prediction sentences using a specified BERT model.

        Args:
            references (Union[str, List[str]]): A single reference sentence or a list of reference sentences.
            predictions (Union[str, List[str]]): A single prediction sentence or a list of prediction sentences.
            model (Optional[str], optional): The name of the BERT model to be used for scoring. Defaults to "microsoft/deberta-large-mnli".
            lang (Optional[str], optional): The language code of the text, e.g., "en" for English. Defaults to "en".

        Returns:
            Dict[str, float]: A dictionary containing BERTScore metrics including precision, recall, and F1 score.
                - 'bert-precision' (float): BERTScore precision.
                - 'bert-recall' (float): BERTScore recall.
                - 'bert-f1' (float): BERTScore F1 score.

        Note:
            Before using this function, make sure to install the 'bert_score' module by running the following command:
            ```
            pip install bert-score
            ```
        """
        try:
            from bert_score import BERTScorer
        except ModuleNotFoundError as e:
            print(
                "Please install bert_score module. Command: pip install bert-score"
            )

        # FIXME: Fix the case for mps
        device = "cuda" if torch.cuda.is_available() else "cpu"
        bert_scorer = BERTScorer(
            model_type=model,
            lang=lang,
            rescale_with_baseline=True,
            device=device,
        )

        if isinstance(predictions, str):
            predictions = [predictions]

        if isinstance(references, str):
            references = [references]

        if (
            isinstance(predictions, list)
            and isinstance(references, list)
            and not isinstance(references[0], list)
        ):
            if len(predictions) != len(references):
                references = [references]

        precision, recall, f1 = bert_scorer.score(
            cands=predictions, refs=references
        )
        return {
            "bert-precision": precision.detach().numpy().tolist(),
            "bert-recall": recall.detach().numpy().tolist(),
            "bert-f1": f1.detach().numpy().tolist(),
        }

    @classmethod
    def faithfulness_score(
        cls, target: str, prediction: str, model: Optional[str] = None
    ) -> float:
        """Calculate the faithfulness score of a prediction compared to a target text using SummaCZS.

        This method computes a faithfulness score, which measures the extent to which a generated prediction matches the provided target text.
        The score is based on the SummaCZS (Summarization Competence with Zero-shot Supervision) model.

        Args:
            target (str): The reference target text for comparison.
            prediction (str): The generated prediction to be evaluated.
            model (Optional[str], optional): The SummaCZS model name to use. If not provided, the "vitc" model will be used by default.

        Returns:
            float: The computed faithfulness score. Higher values indicate greater faithfulness to the target text.
        """
        model = "vitc" if model is None else model
        device = "cuda" if torch.cuda.is_available() else "cpu"
        scorer = SummaCZS(
            granularity="sentence",
            model_name=model,
            imager_load_cache=False,
            device=device,
        )
        return scorer.score_one(target, prediction)["score"]

    @classmethod
    def hallucination_score(
        cls, source: str, prediction: str, model: Optional[str] = None
    ) -> float:
        """Calculate the hallucination score of a prediction compared to a source text.

        This method computes a hallucination score, which measures the extent to which a generated prediction contains hallucinations.
        The score is based on the Vectara Hallucination Evaluation Model.

        Args:
            source (str): The source document where the information is summarized from.
            prediction (str): The generated summary that is validated against the source summary.

        Returns:
            float: The computed hallucination score. Lower values indicate greater hallucination.
        """
        try:
            from deepeval.models.hallucination_model import (
                HallucinationModel,
            )
        except ImportError as e:
            print(e)
        model = "vectara-hallucination" if model is None else model

        scorer = HallucinationModel(model_name=model)

        return scorer.model.predict([source, prediction])

    @classmethod
    def PII_score(
        cls, target: str, prediction: str, model: Optional[Any] = None
    ) -> float:
        raise NotImplementedError()

    @classmethod
    def neural_toxic_score(
        cls, prediction: str, model: Optional[Any] = None
    ) -> Union[float, dict]:
        """
        Calculate the toxicity score of a given text prediction using the Detoxify model.

        Args:
            prediction (str): The text prediction to evaluate for toxicity.
            model (Optional[str], optional): The variant of the Detoxify model to use.
                Available variants: 'original', 'unbiased', 'multilingual'.
                If not provided, the 'original' variant is used by default.

        Returns:
            Union[float, dict]: The mean toxicity score, ranging from 0 (non-toxic) to 1 (highly toxic),
            and also a dictionary containing different types of toxicity score.

        For each model, we get mean toxicity score and a dictionary containing different toxicity score types.
        Examples:
        If model is 'original', we get the a dict with the following keys:
            - 'toxicity',
            - 'severe_toxicity',
            - 'obscene',
            - 'threat'
            - 'insult'
            - 'identity_attack'

        If model is 'unbiased', we get a dict with the same as keys as 'original', but
        along with `sexual_explicit`.

        If the model is 'multilingual', we get a dict same as the unbiasd one.
        """
        try:
            from detoxify import Detoxify
        except ImportError as e:
            print(e)

        device = "cuda" if torch.cuda.is_available() else "cpu"
        if model is not None:
            assert model in [
                "original",
                "unbiased",
                "multilingual",
            ], "Invalid model. Available variants: original, unbiased, multilingual"
            detoxify_model = Detoxify(model, device=device)
        else:
            detoxify_model = Detoxify("original", device=device)
        toxicity_score_dict = detoxify_model.predict(prediction)
        mean_toxicity_score = sum(list(toxicity_score_dict.values())) / len(
            toxicity_score_dict
        )
        return mean_toxicity_score, toxicity_score_dict



import pytest
import openai
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.metrics import LLMEvalMetric
from deepeval.evaluator import assert_test


def test_chat_completion():
    """Test Chat Completion"""
    metric = LLMEvalMetric(
        name="Validity",
        criteria="The response is a valid response to the prompt.",
        minimum_score=0.5,
        evaluation_params=[
            LLMTestCaseParams.INPUT,
            LLMTestCaseParams.ACTUAL_OUTPUT,
        ],
    )
    test_case = LLMTestCase(
        input="What is the capital of France?",
        actual_output="Paris",
        expected_output="Paris",
        context=["Geography"],
    )

    assert_test(test_case, [metric])


# # set openai api type
# openai.api_type = "azure"

# # The azure openai version you want to use
# openai.api_version = "2023-03-15"

# # The base URL for your Azure OpenAI resource.  You can find this in the Azure portal under your Azure OpenAI resource.
# openai.api_base = "https://your-resource-name.openai.azure.com/"
# openai.api_key = "<your Azure OpenAI API key>"

# def test_azure_openai_chat_completion():
#     """Test Chat Completion"""
#     metric = LLMEvalMetric(
#         name="Validity",
#         criteria="The response is a valid response to the prompt.",
#         minimum_score=0.5,
#         evaluation_params=[
#             LLMTestCaseParams.INPUT,
#             LLMTestCaseParams.ACTUAL_OUTPUT,
#         ],
#         deployment_id="your-deployment-id",
#     )
#     test_case = LLMTestCase(
#         input="What is the capital of France?",
#         actual_output="Paris",
#         expected_output="Paris",
#         context=["Geography"],
#     )

#     assert_test(test_case, [metric])



import pytest
from deepeval.test_case import LLMTestCase
from deepeval.metrics import (
    RagasMetric,
    ContextualRelevancyMetric,
    FaithfulnessMetric,
    ContextRecallMetric,
    ConcisenessMetric,
    CorrectnessMetric,
    CoherenceMetric,
    MaliciousnessMetric,
)
from deepeval.metrics.ragas_metric import AnswerRelevancyMetric
from deepeval.evaluator import assert_test

query = "Who won the FIFA World Cup in 2018?"
output = "Winners of the FIFA world cup were the French national football team"
expected_output = "French national football team"
context = [
    "The FIFA World Cup in 2018 was won by the French national football team. They defeated Croatia 4-2 in the final match to claim the championship."
]


@pytest.mark.skip(reason="openai is expensive")
def test_ragas_score():
    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        context=context,
    )
    metric = RagasMetric()

    with pytest.raises(AssertionError):
        assert_test(
            test_case=[test_case],
            metrics=[metric],
        )


@pytest.mark.skip(reason="openai is expensive")
def test_everything():
    test_case = LLMTestCase(
        input=query,
        actual_output=output,
        expected_output=expected_output,
        context=context,
    )
    metric1 = ContextualRelevancyMetric()
    metric2 = FaithfulnessMetric()
    metric3 = ContextRecallMetric()
    metric4 = ConcisenessMetric()
    metric5 = CorrectnessMetric()
    metric6 = CoherenceMetric()
    metric7 = MaliciousnessMetric()
    metric8 = AnswerRelevancyMetric()
    assert_test(
        test_case,
        [
            metric1,
            metric2,
            metric3,
            metric4,
            metric5,
            metric6,
            metric7,
            metric8,
        ],
    )



import os
from deepeval.singleton import Singleton
from deepeval.test_case import LLMTestCase
from deepeval.utils import chunk_text, softmax
from deepeval.metrics.base_metric import BaseMetric
from deepeval.progress_context import progress_context
from sentence_transformers import CrossEncoder


class FactualConsistencyModel(metaclass=Singleton):
    def __init__(self, model_name: str = "cross-encoder/nli-deberta-v3-large"):
        # We use a smple cross encoder model
        os.environ["TOKENIZERS_PARALLELISM"] = "false"
        self.model = CrossEncoder(model_name)

    def predict(self, text_a: str, text_b: str):
        scores = self.model.predict([(text_a, text_b), (text_b, text_a)])
        # https://huggingface.co/cross-encoder/nli-deberta-base
        # label_mapping = ["contradiction", "entailment", "neutral"]
        softmax_scores = softmax(scores)
        score = softmax_scores[0][1]
        second_score = softmax_scores[1][1]
        return max(score, second_score)


class FactualConsistencyMetric(BaseMetric, metaclass=Singleton):
    def __init__(
        self,
        minimum_score: float = 0.6,
        model_name: str = "cross-encoder/nli-deberta-v3-large",
    ):
        # For Crossencoder model, move to singleton to avoid re-instantiating

        with progress_context(
            "Downloading FactualConsistencyModel (may take up to 2 minutes if running for the first time)..."
        ):
            self.model = FactualConsistencyModel(model_name)
        self.minimum_score = minimum_score

    def measure(self, test_case: LLMTestCase):
        if test_case.actual_output is None or test_case.context is None:
            raise ValueError("Output or context cannot be None")

        context_list = []
        if isinstance(test_case.context, str):
            context_list.extend(chunk_text(test_case.context))
        elif isinstance(test_case.context, list):
            for context in test_case.context:
                context_list.extend(chunk_text(context))
        else:
            raise ValueError("Context must be a string or a list of strings")

        max_score = 0
        for c in context_list:
            score = self.model.predict(c, test_case.actual_output)
            if score > max_score:
                max_score = score

        self.success = max_score > self.minimum_score
        self.score = max_score
        return max_score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Factual Consistency"



import os
from deepeval.singleton import Singleton
from sentence_transformers import CrossEncoder
from deepeval.progress_context import progress_context
from deepeval.models.model_map import model_map, name_to_card


class HallucinationModel(metaclass=Singleton):
    def __init__(self, model_name: str = "vectara-hallucination"):
        # We use a smple cross encoder model
        os.environ["TOKENIZERS_PARALLELISM"] = "false"

        # TODO: add this progress context in the correct place
        with progress_context(
            "Downloading HallucinationEvaluationModel (may take up to 2 minutes if running for the first time)..."
        ):
            model_name = name_to_card(model_name)
            self.model = CrossEncoder(model_name)



import os

from ..constants import API_KEY_ENV
from ..key_handler import KEY_FILE_HANDLER


def set_api_key():
    api_key = KEY_FILE_HANDLER.fetch_api_key()
    os.environ[API_KEY_ENV] = api_key


def set_env_vars():
    set_api_key()



"""Investigate test case.
"""
import hashlib
from dataclasses import dataclass
from typing import List, Optional
from enum import Enum


class LLMTestCaseParams(Enum):
    INPUT = "input"
    ACTUAL_OUTPUT = "actual_output"
    EXPECTED_OUTPUT = "expected_output"
    CONTEXT = "context"


@dataclass
class TestCase:
    id: Optional[str] = None

    def __post_init__(self):
        if self.id is None:
            id_string = "".join(str(value) for value in self.__dict__.values())
            self.id = hashlib.md5(id_string.encode()).hexdigest()


@dataclass
class LLMTestCase(TestCase):
    def __init__(
        self,
        input: str,
        actual_output: str,
        expected_output: Optional[str] = None,
        context: Optional[List[str]] = None,
        id: Optional[str] = None,
    ):
        super().__init__(id)
        self.input = input
        self.actual_output = actual_output
        self.expected_output = expected_output
        self.context = context

    def __post_init__(self):
        super().__post_init__()
        self.__name__ = f"LLMTestCase_{self.id}"


class AgentTestCase(TestCase):
    """Test Case For Agents"""

    pass



"""Test for custom metrics in Python
"""

from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.evaluator import assert_test


class LengthMetric(BaseMetric):
    """This metric checks if the output is more than 3 letters"""

    def __init__(self, minimum_score: int = 3):
        self.minimum_score = minimum_score

    def measure(self, test_case: LLMTestCase):
        # sends to server
        text = test_case.actual_output
        score = len(text)
        self.success = score > self.minimum_score
        # Optional: Logs it to the server
        return score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Length"


def test_length_metric():
    metric = LengthMetric()
    test_case = LLMTestCase(
        input="placeholder",
        actual_output="This is a long sentence that is more than 3 letters",
    )
    assert_test(test_case, [metric])



# mypy: check_untyped_defs = False
###############################################
# Source: https://github.com/tingofurro/summac
###############################################

from transformers import AutoTokenizer, AutoModelForSequenceClassification
import nltk
import numpy as np
import torch
import os
import json
from deepeval import utils as utils_misc
from deepeval.models.model_map import name_to_card, get_neutral_idx, model_map


class SummaCImager:
    def __init__(
        self,
        model_name="mnli",
        granularity="paragraph",
        use_cache=True,
        max_doc_sents=100,
        device="cuda",
        **kwargs
    ):
        self.grans = granularity.split("-")

        assert (
            all(
                gran in ["paragraph", "sentence", "document", "2sents", "mixed"]
                for gran in self.grans
            )
            and len(self.grans) <= 2
        ), "Unrecognized `granularity` %s" % (granularity)
        assert (
            model_name in model_map.keys()
        ), "Unrecognized model name: `%s`" % (model_name)

        self.model_name = model_name
        if model_name != "decomp":
            self.model_card = name_to_card(model_name)
            self.entailment_idx = model_map[model_name]["entailment_idx"]
            self.contradiction_idx = model_map[model_name]["contradiction_idx"]
            self.neutral_idx = get_neutral_idx(
                self.entailment_idx, self.contradiction_idx
            )

        self.granularity = granularity
        self.use_cache = use_cache
        self.cache_folder = "/export/share/plaban/summac_cache/"

        self.max_doc_sents = max_doc_sents
        self.max_input_length = 500
        self.device = device
        self.cache = {}
        self.model = None  # Lazy loader

    def load_nli(self):
        if self.model_name == "decomp":
            try:
                from allennlp.predictors.predictor import Predictor
            except ModuleNotFoundError:
                print(
                    "allennlp library is not installed. "
                    "Please install the library by following the instruction from their documentation:"
                    "https://docs.allennlp.org/main/"
                )
            self.model = Predictor.from_path(
                "https://storage.googleapis.com/allennlp-public-models/decomposable-attention-elmo-2020.04.09.tar.gz",
                cuda_device=0,
            )

        else:
            self.tokenizer = AutoTokenizer.from_pretrained(self.model_card)
            self.model = AutoModelForSequenceClassification.from_pretrained(
                self.model_card
            ).eval()
            self.model.to(self.device)

    def split_sentences(self, text):
        sentences = nltk.tokenize.sent_tokenize(text)
        sentences = [sent for sent in sentences if len(sent) > 10]
        return sentences

    def split_2sents(self, text):
        sentences = nltk.tokenize.sent_tokenize(text)
        sentences = [sent for sent in sentences if len(sent) > 10]
        two_sents = [
            " ".join(sentences[i : (i + 2)]) for i in range(len(sentences))
        ]
        return two_sents

    def split_paragraphs(self, text):
        if text.count("\n\n") > 0:
            paragraphs = [p.strip() for p in text.split("\n\n")]
        else:
            paragraphs = [p.strip() for p in text.split("\n")]
        return [p for p in paragraphs if len(p) > 10]

    def split_text(self, text, granularity="sentence"):
        if granularity == "document":
            return [text]
        elif granularity == "paragraph":
            return self.split_paragraphs(text)
        elif granularity == "sentence":
            return self.split_sentences(text)
        elif granularity == "2sents":
            return self.split_2sents(text)
        elif granularity == "mixed":
            return self.split_sentences(text) + self.split_paragraphs(text)

    def build_image(self, original, generated):
        cache_key = (original, generated)
        if self.use_cache and cache_key in self.cache:
            cached_image = self.cache[cache_key]
            cached_image = cached_image[:, : self.max_doc_sents, :]
            return cached_image

        if len(self.grans) == 1:
            gran_doc, gran_sum = self.grans[0], self.grans[0]
        else:
            gran_doc, gran_sum = self.grans[0], self.grans[1]

        original_chunks = self.split_text(original, granularity=gran_doc)[
            : self.max_doc_sents
        ]
        generated_chunks = self.split_text(generated, granularity=gran_sum)

        N_ori = len(original_chunks)
        N_gen = len(generated_chunks)

        if N_ori == 0 or N_gen == 0:
            return np.zeros((3, 1, 1))
        # assert (N_ori > 0 and N_gen > 0), "One of the inputs has no chunks"

        image = np.zeros((3, N_ori, N_gen))

        if self.model is None:
            self.load_nli()

        dataset = [
            {
                "premise": original_chunks[i],
                "hypothesis": generated_chunks[j],
                "doc_i": i,
                "gen_i": j,
            }
            for i in range(N_ori)
            for j in range(N_gen)
        ]
        for batch in utils_misc.batcher(dataset, batch_size=20):
            if self.model_name == "decomp":
                batch_evids, batch_conts, batch_neuts = [], [], []
                batch_json = [
                    {"premise": d["premise"], "hypothesis": d["hypothesis"]}
                    for d in batch
                ]
                model_outs = self.model.predict_batch_json(batch_json)
                for out in model_outs:
                    probs = out["label_probs"]
                    batch_evids.append(probs[0])
                    batch_conts.append(probs[1])
                    batch_neuts.append(probs[2])

            else:
                batch_prems = [b["premise"] for b in batch]
                batch_hypos = [b["hypothesis"] for b in batch]
                batch_tokens = self.tokenizer.batch_encode_plus(
                    list(zip(batch_prems, batch_hypos)),
                    padding=True,
                    truncation=True,
                    max_length=self.max_input_length,
                    return_tensors="pt",
                    truncation_strategy="only_first",
                )
                batch_tokens = {
                    k: v.to(self.device) for k, v in batch_tokens.items()
                }
                with torch.no_grad():
                    model_outputs = self.model(**batch_tokens)

                batch_probs = torch.nn.functional.softmax(
                    model_outputs["logits"], dim=-1
                )
                batch_evids = batch_probs[:, self.entailment_idx].tolist()
                batch_conts = batch_probs[:, self.contradiction_idx].tolist()
                batch_neuts = batch_probs[:, self.neutral_idx].tolist()

            for b, evid, cont, neut in zip(
                batch, batch_evids, batch_conts, batch_neuts
            ):
                image[0, b["doc_i"], b["gen_i"]] = evid
                image[1, b["doc_i"], b["gen_i"]] = cont
                image[2, b["doc_i"], b["gen_i"]] = neut

        if self.use_cache:
            self.cache[cache_key] = image
        return image

    def get_cache_file(self):
        return os.path.join(
            self.cache_folder,
            "cache_%s_%s.json" % (self.model_name, self.granularity),
        )

    def save_cache(self):
        cache_cp = {"[///]".join(k): v.tolist() for k, v in self.cache.items()}
        with open(self.get_cache_file(), "w") as f:
            json.dump(cache_cp, f)

    def load_cache(self):
        cache_file = self.get_cache_file()
        if os.path.isfile(cache_file):
            with open(cache_file, "r") as f:
                cache_cp = json.load(f)
                self.cache = {
                    tuple(k.split("[///]")): np.array(v)
                    for k, v in cache_cp.items()
                }


class SummaCConv(torch.nn.Module):
    def __init__(
        self,
        models=["mnli", "anli", "vitc"],
        bins="even50",
        granularity="sentence",
        nli_labels="e",
        device="cuda",
        start_file=None,
        imager_load_cache=True,
        agg="mean",
        norm_histo=False,
        **kwargs
    ):
        # `bins` should be `even%d` or `percentiles`
        assert nli_labels in [
            "e",
            "c",
            "n",
            "ec",
            "en",
            "cn",
            "ecn",
        ], "Unrecognized nli_labels argument %s" % (nli_labels)

        super(SummaCConv, self).__init__()
        self.device = device
        self.models = models

        self.imagers = []
        for model_name in models:
            self.imagers.append(
                SummaCImager(
                    model_name=model_name, granularity=granularity, **kwargs
                )
            )
        if imager_load_cache:
            for imager in self.imagers:
                imager.load_cache()
        assert len(self.imagers) > 0, "Imager names were empty or unrecognized"

        if "even" in bins:
            n_bins = int(bins.replace("even", ""))
            self.bins = list(np.arange(0, 1, 1 / n_bins)) + [1.0]
        elif bins == "percentile":
            self.bins = [
                0.0,
                0.01,
                0.02,
                0.03,
                0.04,
                0.07,
                0.13,
                0.37,
                0.90,
                0.91,
                0.92,
                0.93,
                0.94,
                0.95,
                0.955,
                0.96,
                0.965,
                0.97,
                0.975,
                0.98,
                0.985,
                0.99,
                0.995,
                1.0,
            ]

        self.nli_labels = nli_labels
        self.n_bins = len(self.bins) - 1
        self.norm_histo = norm_histo
        self.n_rows = 10
        self.n_labels = 2
        self.n_depth = len(self.imagers) * len(self.nli_labels)
        self.full_size = self.n_depth * self.n_bins
        if self.norm_histo:
            self.full_size += (
                2  # Will explicitely give the count of originals and generateds
            )

        self.agg = agg

        self.mlp = torch.nn.Linear(self.full_size, 1).to(device)
        self.layer_final = torch.nn.Linear(3, self.n_labels).to(device)

        if start_file is not None:
            print(self.load_state_dict(torch.load(start_file)))

    def build_image(self, original, generated):
        images = [
            imager.build_image(original, generated) for imager in self.imagers
        ]
        image = np.concatenate(images, axis=0)
        return image

    def compute_histogram(self, original=None, generated=None, image=None):
        # Takes the two texts, and generates a (n_rows, 2*n_bins)

        if image is None:
            image = self.build_image(original, generated)

        N_depth, N_ori, N_gen = image.shape

        full_histogram = []
        for i_gen in range(N_gen):
            histos = []

            for i_depth in range(N_depth):
                if (
                    (i_depth % 3 == 0 and "e" in self.nli_labels)
                    or (i_depth % 3 == 1 and "c" in self.nli_labels)
                    or (i_depth % 3 == 2 and "n" in self.nli_labels)
                ):
                    histo, X = np.histogram(
                        image[i_depth, :, i_gen],
                        range=(0, 1),
                        bins=self.bins,
                        density=self.norm_histo,
                    )
                    histos.append(histo)

            if self.norm_histo:
                histos = [[N_ori, N_gen]] + histos
            histogram_row = np.concatenate(histos)
            full_histogram.append(histogram_row)

        n_rows_missing = self.n_rows - len(full_histogram)
        full_histogram += [[0.0] * self.full_size] * n_rows_missing
        full_histogram = full_histogram[: self.n_rows]
        full_histogram = np.array(full_histogram)
        return image, full_histogram

    def forward(self, originals, generateds, images=None):
        if images is not None:
            # In case they've been pre-computed.
            histograms = []
            for image in images:
                _, histogram = self.compute_histogram(image=image)
                histograms.append(histogram)
        else:
            images, histograms = [], []
            for original, generated in zip(originals, generateds):
                image, histogram = self.compute_histogram(
                    original=original, generated=generated
                )
                images.append(image)
                histograms.append(histogram)

        N = len(histograms)
        histograms = torch.FloatTensor(histograms).to(self.device)

        non_zeros = (torch.sum(histograms, dim=-1) != 0.0).long()
        seq_lengths = non_zeros.sum(dim=-1).tolist()

        mlp_outs = self.mlp(histograms).reshape(N, self.n_rows)
        features = []

        for mlp_out, seq_length in zip(mlp_outs, seq_lengths):
            if seq_length > 0:
                Rs = mlp_out[:seq_length]
                if self.agg == "mean":
                    features.append(
                        torch.cat(
                            [
                                torch.mean(Rs).unsqueeze(0),
                                torch.mean(Rs).unsqueeze(0),
                                torch.mean(Rs).unsqueeze(0),
                            ]
                        ).unsqueeze(0)
                    )
                elif self.agg == "min":
                    features.append(
                        torch.cat(
                            [
                                torch.min(Rs).unsqueeze(0),
                                torch.min(Rs).unsqueeze(0),
                                torch.min(Rs).unsqueeze(0),
                            ]
                        ).unsqueeze(0)
                    )
                elif self.agg == "max":
                    features.append(
                        torch.cat(
                            [
                                torch.max(Rs).unsqueeze(0),
                                torch.max(Rs).unsqueeze(0),
                                torch.max(Rs).unsqueeze(0),
                            ]
                        ).unsqueeze(0)
                    )
                elif self.agg == "all":
                    features.append(
                        torch.cat(
                            [
                                torch.min(Rs).unsqueeze(0),
                                torch.mean(Rs).unsqueeze(0),
                                torch.max(Rs).unsqueeze(0),
                            ]
                        ).unsqueeze(0)
                    )
            else:
                features.append(
                    torch.FloatTensor([0.0, 0.0, 0.0]).unsqueeze(0)
                )  # .cuda()
        features = torch.cat(features)
        logits = self.layer_final(features)
        histograms_out = [histogram.cpu().numpy() for histogram in histograms]
        return logits, histograms_out, images

    def save_imager_cache(self):
        for imager in self.imagers:
            imager.save_cache()

    def score(self, originals, generateds, **kwargs):
        with torch.no_grad():
            logits, histograms, images = self.forward(originals, generateds)
            probs = torch.nn.functional.softmax(logits, dim=-1)
            batch_scores = probs[:, 1].tolist()
        return {
            "scores": batch_scores
        }  # , "histograms": histograms, "images": images


class SummaCZS:
    def __init__(
        self,
        model_name="mnli",
        granularity="paragraph",
        op1="max",
        op2="mean",
        use_ent=True,
        use_con=True,
        imager_load_cache=True,
        device="cuda",
        **kwargs
    ):
        assert op2 in ["min", "mean", "max"], "Unrecognized `op2`"
        assert op1 in ["max", "mean", "min"], "Unrecognized `op1`"

        self.imager = SummaCImager(
            model_name=model_name,
            granularity=granularity,
            device=device,
            **kwargs
        )
        if imager_load_cache:
            self.imager.load_cache()
        self.op2 = op2
        self.op1 = op1
        self.use_ent = use_ent
        self.use_con = use_con

    def save_imager_cache(self):
        self.imager.save_cache()

    def score_one(self, original, generated):
        image = self.imager.build_image(original, generated)

        ent_scores = np.max(image[0], axis=0)
        co_scores = np.max(image[1], axis=0)
        if self.op1 == "mean":
            ent_scores = np.mean(image[0], axis=0)
            co_scores = np.mean(image[1], axis=0)
        elif self.op1 == "min":
            ent_scores = np.min(image[0], axis=0)
            co_scores = np.min(image[1], axis=0)

        if self.use_ent and self.use_con:
            scores = ent_scores - co_scores
        elif self.use_ent:
            scores = ent_scores
        elif self.use_con:
            scores = 1.0 - co_scores

        final_score = np.mean(scores)
        if self.op2 == "min":
            final_score = np.min(scores)
        elif self.op2 == "max":
            final_score = np.max(scores)

        return {"score": final_score, "image": image}

    def score(self, sources, generateds, **kwargs):
        output = {"scores": [], "images": []}
        for source, gen in zip(sources, generateds):
            score = self.score_one(source, gen)
            output["scores"].append(score["score"])
            output["images"].append(score["image"])
        return output



import os
import json
from pydantic import BaseModel, Field
from typing import Any, Optional, List, Dict
from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase
from collections import defaultdict
from deepeval.tracing import get_trace_stack
from deepeval.constants import PYTEST_RUN_TEST_NAME
from deepeval.decorators.hyperparameters import get_hyperparameters
from deepeval.api import Api, Endpoints
import shutil
import webbrowser
from deepeval.utils import delete_file_if_exists
import sys
import datetime
import portalocker
from rich.table import Table
from rich.console import Console
from rich import print

TEMP_FILE_NAME = "temp_test_run_data.json"


class MetricsMetadata(BaseModel):
    metric: str
    score: float
    minimum_score: float = Field(None, alias="minimumScore")
    reason: Optional[str] = None


class MetricScore(BaseModel):
    metric: str
    score: float

    @classmethod
    def from_metric(cls, metric: BaseMetric):
        return cls(metric=metric.__name__, score=metric.score)


class MetricDict:
    def __init__(self):
        self.metric_dict = {}
        self.metric_count = {}

    def add_metric(self, metric_name, score):
        if metric_name not in self.metric_dict:
            self.metric_dict[metric_name] = score
            self.metric_count[metric_name] = 1
        else:
            self.metric_dict[metric_name] += score
            self.metric_count[metric_name] += 1

    def get_average_metric_score(self):
        return [
            MetricScore(
                metric=metric,
                score=self.metric_dict[metric] / self.metric_count[metric],
            )
            for metric in self.metric_dict
        ]


class MetricsMetadataAverageDict:
    def __init__(self):
        self.metric_scores_dict = defaultdict(list)
        self.min_score_dict = defaultdict(float)
        self.metric_reason_dict = defaultdict(str)

    def add_metric(self, metric: BaseMetric):
        metric_name = metric.__name__

        self.metric_scores_dict[metric_name].append(metric.score)
        self.min_score_dict[metric_name] = min(
            self.min_score_dict.get(metric_name, float("inf")),
            metric.minimum_score,
        )
        self.metric_reason_dict[metric_name] = metric.reason

    def get_metrics_metadata(self):
        return [
            MetricsMetadata(
                metric=metric_name,
                score=sum(scores) / len(scores),
                minimumScore=self.min_score_dict[metric_name],
                reason=self.metric_reason_dict[metric_name],
            )
            for metric_name, scores in self.metric_scores_dict.items()
        ]


class APITestCase(BaseModel):
    name: str
    input: str
    actual_output: str = Field(..., alias="actualOutput")
    expected_output: Optional[str] = Field(None, alias="expectedOutput")
    success: bool
    metrics_metadata: List[MetricsMetadata] = Field(
        ..., alias="metricsMetadata"
    )
    run_duration: float = Field(..., alias="runDuration")
    traceStack: Optional[dict] = Field(None)
    context: Optional[list] = Field(None)


class TestRun(BaseModel):
    test_file: Optional[str] = Field(
        None,
        alias="testFile",
    )
    dict_test_cases: Dict[int, APITestCase] = Field(
        default_factory=dict,
    )
    test_cases: List[APITestCase] = Field(
        alias="testCases", default_factory=lambda: []
    )

    metric_scores: List[MetricScore] = Field(
        default_factory=lambda: [], alias="metricScores"
    )
    configurations: Optional[dict[Any, Any]] = Field(default_factory=dict)

    def add_llm_test_case(
        self,
        test_case: LLMTestCase,
        metrics: List[BaseMetric],
        run_duration: float,
        index: int,
    ):
        # Check if test case with the same ID already exists
        test_case_id = id(test_case)
        existing_test_case: LLMTestCase = self.dict_test_cases.get(
            test_case_id, None
        )
        metrics_metadata_dict = MetricsMetadataAverageDict()
        for metric in metrics:
            metrics_metadata_dict.add_metric(metric)
        metrics_metadata = metrics_metadata_dict.get_metrics_metadata()

        if existing_test_case:
            # If it exists, append the metrics to the existing test case
            existing_test_case.metrics_metadata.extend(metrics_metadata)
            success = all(
                [
                    metric.score >= metric.minimum_score
                    for metric in existing_test_case.metrics_metadata
                ]
            )
            # Update the success status
            existing_test_case.success = success
        else:
            # If it doesn't exist, create a new test case
            # Adding backwards compatibility to ensure context still works.
            success = all([metric.is_successful() for metric in metrics])
            api_test_case: APITestCase = APITestCase(
                name=os.getenv(PYTEST_RUN_TEST_NAME, f"test_case_{index}"),
                input=test_case.input,
                actualOutput=test_case.actual_output,
                expectedOutput=test_case.expected_output,
                success=success,
                metricsMetadata=metrics_metadata,
                runDuration=run_duration,
                context=test_case.context,
                traceStack=get_trace_stack(),
            )

            self.dict_test_cases[test_case_id] = api_test_case

    def cleanup(self):
        for _, test_case in self.dict_test_cases.items():
            self.test_cases.append(test_case)
        del self.dict_test_cases
        all_metric_dict = MetricDict()
        for test_case in self.test_cases:
            for metric in test_case.metrics_metadata:
                all_metric_dict.add_metric(metric.metric, metric.score)
        self.metric_scores = all_metric_dict.get_average_metric_score()
        self.configurations = get_hyperparameters()

    def save(self, f):
        json.dump(self.dict(by_alias=True, exclude_none=True), f)
        return self

    @classmethod
    def load(cls, f):
        return cls(**json.load(f))


class TestRunHttpResponse(BaseModel):
    testRunId: str
    projectId: str
    link: str


class TestRunManager:
    def __init__(self):
        self.test_run = None
        self.temp_file_name = TEMP_FILE_NAME
        self.save_to_disk = False

    def set_test_run(self, test_run: TestRun):
        self.test_run = test_run

    def create_test_run(self, file_name: Optional[str] = None):
        test_run = TestRun(
            testFile=file_name,
            testCases=[],
            metricScores=[],
            configurations={},
        )
        self.set_test_run(test_run)

        if self.save_to_disk:
            self.save_test_run()

    def get_test_run(self):
        if self.test_run is None or not self.save_to_disk:
            self.create_test_run()

        if self.save_to_disk:
            try:
                with portalocker.Lock(
                    self.temp_file_name, mode="r", timeout=5
                ) as file:
                    self.test_run = self.test_run.load(file)
            except (FileNotFoundError, portalocker.exceptions.LockException):
                print("Error loading test run from disk", file=sys.stderr)
                self.test_run = None
        return self.test_run

    def save_test_run(self):
        if self.save_to_disk:
            try:
                with portalocker.Lock(
                    self.temp_file_name, mode="w", timeout=5
                ) as file:
                    self.test_run = self.test_run.save(file)
            except portalocker.exceptions.LockException:
                print("Error saving test run to disk", file=sys.stderr)

    def clear_test_run(self):
        self.test_run = None

    # TODO: fix table rendering logic
    def display_results_table(self, test_run: TestRun):
        # Calculate the average of each metric
        metrics_avg = {
            metric.metric: metric.score for metric in test_run.metric_scores
        }

        # Count the number of passes and failures
        # Get all the possible metrics first
        all_metrics = {metric.metric for metric in test_run.metric_scores}

        # Loop through to filter for each metric
        passes = {
            metric: len(
                [
                    test_case_metric
                    for test_case in test_run.test_cases
                    for test_case_metric in test_case.metrics_metadata
                    if test_case_metric.metric == metric and test_case.success
                ]
            )
            for metric in all_metrics
        }
        failures = {
            metric: len(
                [
                    test_case_metric
                    for test_case in test_run.test_cases
                    for test_case_metric in test_case.metrics_metadata
                    if test_case_metric.metric == metric
                ]
            )
            - passes[metric]
            for metric in all_metrics
        }

        table = Table(title="Test Results")
        table.add_column("Metric", justify="right")
        table.add_column("Average Score", justify="right")
        table.add_column("Passes", justify="right")
        table.add_column("Failures", justify="right")
        table.add_column("Success Rate", justify="right")
        total_passes = 0
        total_failures = 0
        for metric, avg in metrics_avg.items():
            pass_count = passes[metric]
            fail_count = failures[metric]
            total_passes += pass_count
            total_failures += fail_count
            success_rate = pass_count / (pass_count + fail_count) * 100
            table.add_row(
                metric,
                str(avg),
                f"[green]{str(pass_count)}[/green]",
                f"[red]{str(fail_count)}[/red]",
                f"{success_rate:.2f}%",
            )
        total_tests = total_passes + total_failures
        overall_success_rate = total_passes / total_tests * 100
        table.add_row(
            "Total",
            "-",
            f"[green]{str(total_passes)}[/green]",
            f"[red]{str(total_failures)}[/red]",
            f"{overall_success_rate:.2f}%",
        )
        print(table)

    def post_test_run(self, test_run: TestRun):
        console = Console()

        # TODO: change this, very hacky way to check if api key exists
        if os.path.exists(".deepeval"):
            try:
                body = test_run.model_dump(by_alias=True, exclude_none=True)
            except AttributeError:
                # Pydantic version below 2.0
                body = test_run.dict(by_alias=True, exclude_none=True)
            api = Api()
            result = api.post_request(
                endpoint=Endpoints.CREATE_TEST_RUN_ENDPOINT.value,
                body=body,
            )
            response = TestRunHttpResponse(
                testRunId=result["testRunId"],
                projectId=result["projectId"],
                link=result["link"],
            )
            if response and os.path.exists(".deepeval"):
                link = response.link
                console.print(
                    "✅ Tests finished! View results on "
                    f"[link={link}]{link}[/link]"
                )
                webbrowser.open(link)
        else:
            console.print(
                '✅ Tests finished! Run "deepeval login" to view evaluation results on the web.'
            )

    def save_test_run_locally(self):
        local_folder = os.getenv("DEEPEVAL_RESULTS_FOLDER")
        if local_folder:
            new_test_filename = datetime.datetime.now().strftime(
                "%Y%m%d_%H%M%S"
            )
            os.rename(self.temp_file_name, new_test_filename)
            if not os.path.exists(local_folder):
                os.mkdir(local_folder)
                shutil.copy(new_test_filename, local_folder)
                print(f"Results saved in {local_folder} as {new_test_filename}")
            elif os.path.isfile(local_folder):
                print(
                    f"""❌ Error: DEEPEVAL_RESULTS_FOLDER={local_folder} already exists and is a file.\nDetailed results won't be saved. Please specify a folder or an available path."""
                )
            else:
                shutil.copy(new_test_filename, local_folder)
                print(f"Results saved in {local_folder} as {new_test_filename}")
            os.remove(new_test_filename)

    def wrap_up_test_run(self, display_table: bool = True):
        test_run = test_run_manager.get_test_run()
        test_run.cleanup()
        if test_run is None or len(test_run.test_cases) == 0:
            print("Test Run is empty, please try again.")
            delete_file_if_exists(test_run_manager.temp_file_name)
            return

        if display_table:
            self.display_results_table(test_run)
        self.post_test_run(test_run)
        self.save_test_run_locally()
        delete_file_if_exists(self.temp_file_name)


test_run_manager = TestRunManager()



"""Metric for bias classifier - using the same min score subtraction methodology as the toxic classifier 
Rationale for bias classifier is described here https://arxiv.org/pdf/2208.05777.pdf
1 - Not Biased
0 - Bias
"""

import warnings
from typing import Optional, List

from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase, LLMTestCaseParams


class UnBiasedMetric(BaseMetric):
    def __init__(
        self,
        evaluation_params: List[LLMTestCaseParams],
        model_name: str = "original",
        minimum_score: float = 0.5,
    ):  # see paper for rationale https://arxiv.org/pdf/2208.05777.pdf
        if not evaluation_params:
            raise ValueError("evaluation_params cannot be empty or None")

        self.evaluation_params = evaluation_params
        self.model_name = model_name
        self.minimum_score = minimum_score

    def __call__(self, output, expected_output, query: Optional[str] = "-"):
        score = self.measure(output, expected_output)
        success = score >= self.minimum_score
        return score

    def measure(self, test_case: LLMTestCase, return_all_scores: bool = False):
        # Check if each param in evaluation_params exists and is not None in test_case
        for param in self.evaluation_params:
            if (
                not hasattr(test_case, param.value)
                or getattr(test_case, param.value) is None
            ):
                raise ValueError(
                    f"Test case is missing the required attribute: {param.value}"
                )

        from Dbias.bias_classification import classifier

        warnings.warn(
            "Run `pip install deepeval[bias]`. If you have, please ignore this warning."
        )

        total_score = 0  # to accumulate scores for all evaluation params
        all_results = (
            []
        )  # to accumulate all individual results if return_all_scores is True

        for param in self.evaluation_params:
            result = classifier(getattr(test_case, param.value))
            if return_all_scores:
                all_results.append(result)

            if result[0]["label"] == "Biased":
                v = 0.5 - (result[0]["score"] / 2)
            else:
                v = 0.5 + (result[0]["score"] / 2)
            total_score += v

        # Calculate the average score
        average_score = total_score / len(self.evaluation_params)

        self.success = average_score > self.minimum_score
        self.score = average_score

        if return_all_scores:
            return all_results

        return average_score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Unbiased Metric"



"""Function for running test
"""

from typing import List
import time
from dataclasses import dataclass
import copy

from deepeval.progress_context import progress_context
from deepeval.metrics import BaseMetric
from deepeval.test_case import LLMTestCase
from deepeval.test_run import test_run_manager


@dataclass
class TestResult:
    """Returned from run_test"""

    success: bool
    metrics: List[BaseMetric]
    input: str
    actual_output: str
    expected_output: str
    context: List[str]


def create_test_result(
    test_case: LLMTestCase,
    success: bool,
    metrics: list[BaseMetric],
) -> TestResult:
    if isinstance(test_case, LLMTestCase):
        return TestResult(
            success=success,
            metrics=metrics,
            input=test_case.input,
            actual_output=test_case.actual_output,
            expected_output=test_case.expected_output,
            context=test_case.context,
        )
    else:
        raise ValueError("TestCase not supported yet.")


def execute_test(
    test_cases: List[LLMTestCase],
    metrics: List[BaseMetric],
    save_to_disk: bool = False,
) -> List[TestResult]:
    test_results: TestResult = []
    test_run_manager.save_to_disk = save_to_disk
    count = 0
    for test_case in test_cases:
        success = True
        for metric in metrics:
            test_start_time = time.perf_counter()
            metric.measure(test_case)
            test_end_time = time.perf_counter()
            run_duration = test_end_time - test_start_time

            test_run_manager.get_test_run().add_llm_test_case(
                test_case=test_case,
                metrics=[metric],
                run_duration=run_duration,
                index=count,
            )
            test_run_manager.save_test_run()

            if not metric.is_successful():
                success = False

        count += 1
        test_result = create_test_result(
            test_case, success, copy.deepcopy(metrics)
        )
        test_results.append(test_result)

    return test_results


def run_test(
    test_case: LLMTestCase,
    metrics: List[BaseMetric],
) -> List[TestResult]:
    with progress_context("Executing run_test()..."):
        test_result = execute_test([test_case], metrics, False)[0]
        print_test_result(test_result)
        print("\n" + "-" * 70)
        return test_result


def assert_test(test_case: LLMTestCase, metrics: List[BaseMetric]):
    # len(execute_test(...)) is always 1 for assert_test
    test_result = execute_test([test_case], metrics, True)[0]
    if not test_result.success:
        failed_metrics = [
            metric
            for metric in test_result.metrics
            if not metric.is_successful()
        ]
        failed_metrics_str = ", ".join(
            [
                f"{metric.__name__} (score: {metric.score}, minimum_score: {metric.minimum_score})"
                for metric in failed_metrics
            ]
        )
        raise AssertionError(f"Metrics {failed_metrics_str} failed.")


def evaluate(test_cases: List[LLMTestCase], metrics: List[BaseMetric]):
    with progress_context("Evaluating testcases..."):
        test_results = execute_test(test_cases, metrics, True)
        for test_result in test_results:
            print_test_result(test_result)
        print("\n" + "-" * 70)

        test_run_manager.wrap_up_test_run(display_table=False)
        return test_results


def print_test_result(test_result: TestResult):
    print("\n" + "=" * 70 + "\n")
    print("Metrics Summary\n")
    for metric in test_result.metrics:
        if not metric.is_successful():
            print(
                f"  - ❌ {metric.__name__} (score: {metric.score}, minimum_score: {metric.minimum_score})"
            )
        else:
            print(
                f"  - ✅ {metric.__name__} (score: {metric.score}, minimum_score: {metric.minimum_score})"
            )

    print("\nFor test case:\n")
    print(f"  - input: {test_result.input}")
    print(f"  - actual output: {test_result.actual_output}")
    print(f"  - expected output: {test_result.expected_output}")
    print(f"  - context: {test_result.context}")



from deepeval.singleton import Singleton
from deepeval.test_case import LLMTestCase
from deepeval.metrics import BaseMetric
from deepeval.scorer import Scorer


class HallucinationMetric(BaseMetric, metaclass=Singleton):
    def __init__(
        self,
        minimum_score: float = 0.5,
    ):
        self.minimum_score = minimum_score

    def measure(self, test_case: LLMTestCase):
        if test_case.actual_output is None or test_case.context is None:
            raise ValueError("Output or context cannot be None")

        context_list = []
        for context in test_case.context:
            context_list.append(context)

        max_score = 0
        for c in context_list:
            score = Scorer.hallucination_score(c, test_case.actual_output)
            if score > max_score:
                max_score = score

        self.success = max_score > self.minimum_score
        self.score = max_score
        return max_score

    def is_successful(self) -> bool:
        return self.success

    @property
    def __name__(self):
        return "Hallucination"



"""Metric for toxic classifier. 
1 - Healthy
0 - Toxic
"""
from typing import List

from deepeval.singleton import Singleton
from deepeval.test_case import LLMTestCase, LLMTestCaseParams
from deepeval.metrics.base_metric import BaseMetric


class DetoxifyModel(metaclass=Singleton):
    def __init__(self, model_name: str = "original"):
        self.model_name = model_name

        try:
            from detoxify import Detoxify
        except ImportError as e:
            print(e)
        self.model = Detoxify(model_name)

    def predict(self, text: str):
        return self.model.predict(text)


class NonToxicMetric(BaseMetric):
    def __init__(
        self,
        evaluation_params: List[LLMTestCaseParams],
        model_name: str = "original",
        minimum_score: float = 0.5,
    ):
        if not evaluation_params:
            raise ValueError("evaluation_params cannot be empty or None")

        self.evaluation_params = evaluation_params
        self.detoxify_model = DetoxifyModel(model_name)
        self.minimum_score = minimum_score

    def __call__(self, test_case: LLMTestCase):
        score = self.measure(test_case.actual_output)
        score = score["min_score"]
        return score

    def measure(self, test_case: LLMTestCase):
        # Check if each param in evaluation_params exists and is not None in test_case
        for param in self.evaluation_params:
            if (
                not hasattr(test_case, param.value)
                or getattr(test_case, param.value) is None
            ):
                raise ValueError(
                    f"Test case is missing the required attribute: {param.value}"
                )

        total_scores = []

        for param in self.evaluation_params:
            text_to_evaluate = getattr(test_case, param.value)
            results = self.detoxify_model.predict(text_to_evaluate)
            # sample output
            # {'toxicity': 0.98057544,
            # 'severe_toxicity': 0.106649496,
            # 'obscene': 0.32923067,
            # 'threat': 0.018646086,
            # 'insult': 0.7514423,
            # 'identity_attack': 0.86643445}

            score = 1 - results["toxicity"]
            total_scores.append(score)

        average_score = sum(total_scores) / len(total_scores)

        # Check if the average score meets the minimum requirement
        self.success = average_score >= self.minimum_score
        self.score = average_score
        return self.score

    def is_successful(self):
        return self.success

    @property
    def __name__(self):
        return "Toxicity"



import os

import pytest
from deepeval.dataset import EvaluationDataset

dataset = EvaluationDataset()


def test_create_dataset():
    module_b_dir = os.path.dirname(os.path.realpath(__file__))

    file_path = os.path.join(module_b_dir, "data", "dataset.csv")

    dataset.add_test_cases_from_csv_file(
        file_path,
        input_col_name="query",
        actual_output_col_name="actual_output",
        expected_output_col_name="expected_output",
        context_col_name="context",
    )
    assert len(dataset.test_cases) == 5, "Test Cases not loaded from CSV"
    file_path = os.path.join(module_b_dir, "data", "dataset.json")
    dataset.add_test_cases_from_json_file(
        file_path,
        input_key_name="query",
        actual_output_key_name="actual_output",
        expected_output_key_name="expected_output",
        context_key_name="context",
    )
    assert len(dataset.test_cases) == 10, "Test Cases not loaded from JSON"



import typer
from typing_extensions import Annotated

# Rich has a few dependency issues
try:
    from rich import print
except Exception as e:
    pass
from deepeval.api import Api
from deepeval.key_handler import KEY_FILE_HANDLER
from deepeval.cli.test import app as test_app
import webbrowser

app = typer.Typer(name="deepeval")

app.add_typer(test_app, name="test")


@app.command()
def login(
    api_key: Annotated[
        str,
        typer.Option(
            help="API key to get from https://app.confident-ai.com. Required if you want to log events to the server."
        ),
    ] = ""
):
    """Login to the DeepEval platform."""
    print("Welcome to [bold]DeepEval[/bold]!")
    print(
        "Grab your API key here: [link=https://app.confident-ai.com]https://app.confident-ai.com[/link] "
    )
    webbrowser.open("https://app.confident-ai.com")
    if api_key == "":
        while True:
            api_key = input("Paste your API Key: ").strip()
            if api_key:
                break
            else:
                print("API Key cannot be empty. Please try again.\n")
    KEY_FILE_HANDLER.write_api_key(api_key)
    client = Api(api_key=api_key)
    print("Congratulations! Login successful :raising_hands: ")
    print(
        "If you are new to DeepEval, follow our quickstart tutorial here: [bold][link=https://docs.confident-ai.com/docs/getting-started]https://docs.confident-ai.com/docs/getting-started[/link][/bold]"
    )


if __name__ == "__main__":
    app()



model_map = {
    "snli-base": {
        "model_card": "boychaboy/SNLI_roberta-base",
        "entailment_idx": 0,
        "contradiction_idx": 2,
    },
    "snli-large": {
        "model_card": "boychaboy/SNLI_roberta-large",
        "entailment_idx": 0,
        "contradiction_idx": 2,
    },
    "mnli-base": {
        "model_card": "microsoft/deberta-base-mnli",
        "entailment_idx": 2,
        "contradiction_idx": 0,
    },
    "mnli": {
        "model_card": "roberta-large-mnli",
        "entailment_idx": 2,
        "contradiction_idx": 0,
    },
    "anli": {
        "model_card": "ynie/roberta-large-snli_mnli_fever_anli_R1_R2_R3-nli",
        "entailment_idx": 0,
        "contradiction_idx": 2,
    },
    "vitc-base": {
        "model_card": "tals/albert-base-vitaminc-mnli",
        "entailment_idx": 0,
        "contradiction_idx": 1,
    },
    "vitc": {
        "model_card": "tals/albert-xlarge-vitaminc-mnli",
        "entailment_idx": 0,
        "contradiction_idx": 1,
    },
    "vitc-only": {
        "model_card": "tals/albert-xlarge-vitaminc",
        "entailment_idx": 0,
        "contradiction_idx": 1,
    },
    # "decomp": 0,
    "vectara-hallucination": {
        "model_card": "vectara/hallucination_evaluation_model",
        "entailment_idx": None,
        "contradiction_idx": None,
    },
}


def card_to_name(card):
    card2name = {v["model_card"]: k for k, v in model_map.items()}
    if card in card2name:
        return card2name[card]
    return card


def name_to_card(name):
    if name in model_map:
        return model_map[name]["model_card"]
    return name


def get_neutral_idx(ent_idx, con_idx):
    return list(set([0, 1, 2]) - set([ent_idx, con_idx]))[0]
